import 'dart:async';

import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';

import 'lyric.dart';
import 'lyric_controller.dart';
import 'lyric_painter.dart';

class LyricWidget extends StatefulWidget {
  final List<Lyric> lyrics;
  final List<Lyric>? remarkLyrics;
  final Size size;
  final LyricController controller;
  final TextStyle? lyricStyle;
  final TextStyle? remarkStyle;
  final TextStyle? currLyricStyle;
  final TextStyle? currRemarkLyricStyle;
  final TextStyle? draggingLyricStyle;
  final TextStyle? draggingRemarkLyricStyle;
  final double lyricGap;
  final double remarkLyricGap;
  final bool enableDrag;

  //字体最大宽度
  final double? lyricMaxWidth;

  LyricWidget(
      {Key? key,
      required this.lyrics,
      this.remarkLyrics,
      required this.size,
      required this.controller,
      this.lyricStyle,
      this.remarkStyle,
      this.currLyricStyle,
      this.lyricGap = 10,
      this.remarkLyricGap = 20,
      this.draggingLyricStyle,
      this.draggingRemarkLyricStyle,
      this.enableDrag: true,
      this.lyricMaxWidth,
      this.currRemarkLyricStyle})
      : super(key: key);

  @override
  State<LyricWidget> createState() => LyricWidgetState();
}

class LyricWidgetState extends State<LyricWidget> {
  LyricPainter? _lyricPainter;
  double totalHeight = 0;

  late List<Lyric> lyrics;
  List<Lyric>? remarkLyrics;
  double? lyricMaxWidth;
  TextStyle? lyricStyle;
  TextStyle? remarkStyle;
  TextStyle? currLyricStyle;
  TextStyle? currRemarkLyricStyle;
  TextStyle? draggingLyricStyle;
  TextStyle? draggingRemarkLyricStyle;

  //歌词画笔数组
  List<TextPainter> lyricTextPaints = [];

  //翻译/音译歌词画笔数组
  List<TextPainter> subLyricTextPaints = [];

  @override
  void initState() {
    super.initState();
    lyrics = widget.lyrics;
    remarkLyrics = widget.remarkLyrics;
    lyricMaxWidth = widget.lyricMaxWidth;
    lyricStyle =
        widget.lyricStyle ?? TextStyle(color: Colors.grey, fontSize: 14);
    remarkStyle =
        widget.remarkStyle ?? TextStyle(color: Colors.black, fontSize: 14);
    currLyricStyle =
        widget.currLyricStyle ?? TextStyle(color: Colors.red, fontSize: 18);
    currRemarkLyricStyle = widget.currRemarkLyricStyle ?? currLyricStyle;
    draggingLyricStyle = widget.draggingLyricStyle ??
        lyricStyle?.copyWith(color: Colors.greenAccent);
    draggingRemarkLyricStyle = widget.draggingRemarkLyricStyle ??
        remarkStyle?.copyWith(color: Colors.greenAccent);
    widget.controller.draggingComplete = () {
      cancelTimer();
      widget.controller.progress =
          widget.controller.draggingProgress ?? Duration();
      _lyricPainter!.draggingLine = null;
      widget.controller.isDragging = false;
    };
    WidgetsBinding.instance!.addPostFrameCallback((call) {
      totalHeight = computeScrollY(lyrics.length - 1);
    });
    widget.controller.addListener(() {
      var curLine =
          findLyricIndexByDuration(widget.controller.progress, lyrics);
      if (widget.controller.oldLine != curLine) {
        _lyricPainter!.currentLyricIndex = curLine;
        if (!widget.controller.isDragging) {
          if (widget.controller.vsync == null) {
            _lyricPainter!.offset = -computeScrollY(curLine);
          } else {
            animationScrollY(curLine, widget.controller.vsync!);
          }
        }
        widget.controller.oldLine = curLine;
      }
    });
    initLyricList();
  }

  @override
  void didUpdateWidget(LyricWidget oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (lyrics != widget.lyrics) {
      lyrics = widget.lyrics;
      remarkLyrics = widget.remarkLyrics;
      cancelTimer();
      widget.controller.progress = Duration();
      widget.controller.draggingProgress = Duration();
      _lyricPainter!.draggingLine = null;
      widget.controller.isDragging = false;
      initLyricList();
    }
  }

  void setLyricList(List<Lyric> lyrics, {List<Lyric>? remarkLyrics}) {
    this.lyrics = lyrics;
    this.remarkLyrics = remarkLyrics;
    initLyricList();
  }
  void initLyricList() {
    lyricTextPaints.clear();
    subLyricTextPaints.clear();
    //歌词转画笔
    lyricTextPaints.addAll(lyrics
        .map((l) => TextPainter(
            text: TextSpan(text: l.lyric, style: lyricStyle),
            textDirection: TextDirection.ltr))
        .toList());

    //翻译/音译歌词转画笔
    if (remarkLyrics != null && remarkLyrics!.isNotEmpty) {
      subLyricTextPaints.addAll(remarkLyrics!
          .map((l) => TextPainter(
              text: TextSpan(text: l.lyric, style: remarkStyle),
              textDirection: TextDirection.ltr))
          .toList());
    }
  }

  ///因空行高度与非空行高度不一致，获取非空行的位置
  int getNotEmptyLineHeight(List<Lyric> lyrics) => lyrics.indexOf(
      lyrics.firstWhere((lyric) => lyric.lyric!.trim().isNotEmpty,
          orElse: () => lyrics.first));

  @override
  Widget build(BuildContext context) {
    if (lyricMaxWidth == null || lyricMaxWidth == double.infinity) {
      lyricMaxWidth = MediaQuery.of(context).size.width;
    }

    _lyricPainter = LyricPainter(lyrics, lyricTextPaints, subLyricTextPaints,
        vsync: widget.controller.vsync,
        subLyrics: remarkLyrics,
        lyricTextStyle: lyricStyle,
        subLyricTextStyle: remarkStyle,
        currLyricTextStyle: currLyricStyle,
        lyricGapValue: widget.lyricGap,
        lyricMaxWidth: lyricMaxWidth,
        subLyricGapValue: widget.remarkLyricGap,
        draggingLyricTextStyle: draggingLyricStyle,
        draggingSubLyricTextStyle: draggingRemarkLyricStyle,
        currSubLyricTextStyle: currRemarkLyricStyle);
    _lyricPainter!.currentLyricIndex =
        findLyricIndexByDuration(widget.controller.progress, lyrics);
    if (widget.controller.isDragging) {
      _lyricPainter!.draggingLine = widget.controller.draggingLine;
      _lyricPainter!.offset = widget.controller.draggingOffset;
    } else {
      _lyricPainter!.offset = -computeScrollY(_lyricPainter!.currentLyricIndex);
    }
    return widget.enableDrag
        ? GestureDetector(
            onVerticalDragUpdate: (e) {
              cancelTimer();
              double temOffset = (_lyricPainter!.offset + e.delta.dy);
              if (temOffset < 0 && temOffset >= -totalHeight) {
                widget.controller.draggingOffset = temOffset;
                widget.controller.draggingLine =
                    getCurrentDraggingLine(temOffset + widget.lyricGap);
                _lyricPainter!.draggingLine = widget.controller.draggingLine;
                widget.controller.draggingProgress =
                    lyrics[widget.controller.draggingLine].startTime! +
                        Duration(milliseconds: 1);
                widget.controller.isDragging = true;
                _lyricPainter!.offset = temOffset;
              }
            },
            onVerticalDragEnd: (e) {
              cancelTimer();
              widget.controller.draggingTimer = Timer(
                  widget.controller.draggingTimerDuration ??
                      Duration(seconds: 3), () {
                resetDragging();
              });
              widget.controller.draggingEnd?.call();
            },
            child: buildCustomPaint(),
          )
        : buildCustomPaint();
  }

  CustomPaint buildCustomPaint() {
    return CustomPaint(
      painter: _lyricPainter,
      size: widget.size,
    );
  }

  void resetDragging() {
    _lyricPainter!.currentLyricIndex =
        findLyricIndexByDuration(widget.controller.progress, lyrics);

    widget.controller.previousRowOffset =
        -(widget.controller.draggingOffset ?? 0);
    animationScrollY(
        _lyricPainter!.currentLyricIndex, widget.controller.vsync!);
    _lyricPainter!.draggingLine = null;
    widget.controller.isDragging = false;
  }

  int getCurrentDraggingLine(double offset) {
    for (int i = 0; i < lyrics.length; i++) {
      var scrollY = computeScrollY(i);
      if (offset > -1) {
        offset = 0;
      }
      if (offset >= -scrollY) {
        return i;
      }
    }
    return lyrics.length;
  }

  void cancelTimer() {
    if (widget.controller.draggingTimer != null) {
      if (widget.controller.draggingTimer!.isActive) {
        widget.controller.draggingTimer!.cancel();
        widget.controller.draggingTimer = null;
      }
    }
  }

  animationScrollY(currentLyricIndex, TickerProvider tickerProvider) {
    var animationController = widget.controller.animationController;
    if (animationController != null) {
      animationController.stop();
    }
    animationController = AnimationController(
        vsync: tickerProvider, duration: Duration(milliseconds: 300))
      ..addStatusListener((status) {
        if (status == AnimationStatus.completed) {
          animationController!.dispose();
          animationController = null;
        }
      });
    // 计算当前行偏移量
    var currentRowOffset = computeScrollY(currentLyricIndex);
    //如果偏移量相同不执行动画
    if (currentRowOffset == widget.controller.previousRowOffset) {
      return;
    }
    // 起始为上一行，结束点为当前行
    Animation animation = Tween<double>(
            begin: widget.controller.previousRowOffset, end: currentRowOffset)
        .animate(animationController!);
    widget.controller.previousRowOffset = currentRowOffset;
    animationController!.addListener(() {
      _lyricPainter!.offset = -animation.value;
    });
    animationController!.forward();
  }

  //根据当前时长获取歌词位置
  int findLyricIndexByDuration(Duration curDuration, List<Lyric> lyrics) {
    for (int i = 0; i < lyrics.length; i++) {
      if (curDuration >= lyrics[i].startTime! &&
          curDuration <= lyrics[i].endTime!) {
        return i;
      }
    }
    return 0;
  }

  /// 计算传入行和第一行的偏移量
  double computeScrollY(int curLine) {
    double totalHeight = 0;
    for (var i = 0; i < curLine; i++) {
      var currPaint = lyricTextPaints[i]
        ..text = TextSpan(text: lyrics[i].lyric, style: lyricStyle);
      currPaint.layout(maxWidth: lyricMaxWidth!);
      totalHeight += currPaint.height + widget.lyricGap;
    }
    if (remarkLyrics != null) {
      //增加 当前行之前的翻译歌词的偏移量
      remarkLyrics!
          .where((subLyric) => subLyric.endTime! <= lyrics[curLine].endTime!)
          .toList()
          .forEach((subLyric) {
        var currentPaint = subLyricTextPaints[remarkLyrics!.indexOf(subLyric)]
          ..text = TextSpan(text: subLyric.lyric, style: remarkStyle);
        currentPaint.layout(maxWidth: lyricMaxWidth!);
        totalHeight += widget.remarkLyricGap + currentPaint.height;
      });
    }
    return totalHeight;
  }
}
